bl_info = {
    'name': 'Artist 3D Viewer animation tool',
    'author': 'Tlakali Studio',
    'version': (0, 1, 0),
    'blender': (2, 80, 0),
    'category': 'Armature',
    'description': 'This script import animations exported from Artist 3D Viewer (MoCap recorded animations)'
}

import bpy
import math
import mathutils
import json

def cmpEqVal(fl_a, fl_b, fl_threshold):
    if abs(fl_a - fl_b) < fl_threshold:
        return True
    else:
        return False

def cmpEqQuats(quat_A, quat_B, fl_threshold):
    if cmpEqVal(quat_A['w'], quat_B['w'], fl_threshold):
        return True
    else:
        return False

def cmpEqVects(vec_A, vec_B, fl_threshold):
    if (vec_A - vec_B).length < fl_threshold:
        return True
    else:
        return False

class WM_OT_animation_loader(bpy.types.Operator):
    """Loads A3DV animation file creating action and bone keyframes"""
    # Operator properties
    bl_idname = "wm.anim_loader"
    bl_label = "Load animation file"
    bl_options = {'REGISTER', 'UNDO'}
    
    def execute(self, context):
        scn = context.scene
        armature = bpy.context.object
        win_man = context.window_manager
        bool_verb = scn.animLoadingVerbose
        bool_overwAct = scn.animOverwriteAct
        bool_applyTras = scn.animApplyTras
        bool_optimize = scn.animOptimize
        fl_optThreshold = scn.animOptimThreshold
        
        fl_optLocThreshold = 0.001
        fl_locScale = 10
        
        # TO_DO : Check empty file
        # ---- Reading JSON File
        animFile = open(bpy.path.abspath(scn.animFilePath))
        animDef = json.load(animFile)
        animFile.close()
        str_action_name = animDef['name']
        
        blenderFPS = bpy.context.scene.render.fps
        a3dvFPS = animDef['framesPerSecond']
        delta = 1
        if a3dvFPS < blenderFPS:
            delta = math.ceil(blenderFPS/a3dvFPS)
        print("++ Processing animation json file ...")
        print("   Action : ", str_action_name)
        print("   Blender FPS : ", blenderFPS)
        print("   A3DV FPS : ", a3dvFPS)
        print("   Delta : ", delta)
        print("   Verbose : ", bool_verb)

        # ---- Initiating progress tracking
        num_maxProg = len(animDef['bonesOrientation'])
        num_deltaProg = 100 / num_maxProg
        win_man.progress_begin(0, num_maxProg)
        print("+ Num. Bones : ", num_maxProg)

        # ---- Creating or replacing action
        armature.animation_data_create()
        try:
            action = bpy.data.actions[str_action_name]
            if bool_overwAct:
                print("++ Action [", str_action_name, "] will be overwritten")
                bpy.data.actions.remove(action)

            armature.animation_data.action = bpy.data.actions.new(name=str_action_name)
            action = bpy.data.actions[str_action_name]
        except KeyError:
            print("++ Action [", str_action_name, "] will be created")
            armature.animation_data.action = bpy.data.actions.new(name=str_action_name)
            action = bpy.data.actions[str_action_name]

        # ---- APPLYING TRANSFORMATIONS
        bpy.ops.object.mode_set(mode='POSE')

        # ---- Applying root bone traslation
        if bool_applyTras:
            str_root_name = "root"
            pbone_root = armature.pose.bones[str_root_name]
            vec_loc_prev = mathutils.Vector((-99.9, -99.9, -99.9))
            vec_loc_curr = mathutils.Vector((0.0, 0.0, 0.0))
            
            arr_positions = animDef['rootPosition']
            int_array_length = len(arr_positions)
            for idx, rootPos in enumerate(arr_positions):
                vec_loc_curr.x = rootPos['x']
                vec_loc_curr.y = rootPos['y']
                vec_loc_curr.z = rootPos['z']
                vec_loc_curr = vec_loc_curr * fl_locScale
                if bool_optimize:
                    if not cmpEqVects(vec_loc_prev, vec_loc_curr, fl_optLocThreshold) or (idx + 1) == int_array_length:
                        vec_loc_prev.x = vec_loc_curr.x
                        vec_loc_prev.y = vec_loc_curr.y
                        vec_loc_prev.z = vec_loc_curr.z
                    else:
                        continue
                
                pbone_root.location = vec_loc_curr.copy()
                pbone_root.keyframe_insert(data_path='location',frame=(idx * delta))

        # ---- Applying bones rotations
        # ---- Iterating over bones array
        int_boneIdx = 0
        int_tmpProg = 0
        print("++ Starting loading animation ...")
        for boneName, boneDef in animDef['bonesOrientation'].items():
            #if boneName == 'modelNode':
            #    boneName = 'root'
                #continue
            if bool_verb:
                print("++ Loading bone keyframes [" + boneName + "]")

            # ---- Getting current pose bone and setting transformation mode
            pbone_current = armature.pose.bones[boneName]
            pbone_current.rotation_mode = 'QUATERNION'
            
            # ---- Previous quaternion
            quat_prev = {"w":0.0, "x":0.0, "y":0.0, "z":0.0}
            quat_curr = {"w":0.0, "x":0.0, "y":0.0, "z":0.0}
            
            # ---- Iterating over orientations array
            for idx, rotDef in enumerate(boneDef):
                # ---- Current bone (from the pose bone)
                bone_temp = pbone_current.bone
                quat_curr['w'] = float(rotDef['w'])
                quat_curr['x'] = float(rotDef['x'])
                quat_curr['y'] = float(rotDef['y'])
                quat_curr['z'] = float(rotDef['z'])
                
                # ---- When optimize flag is on the current transformation will be skipped if is equal than the previous one
                if bool_optimize:
                    if not cmpEqQuats(quat_prev, quat_curr, fl_optThreshold) or (idx + 1) == len(boneDef):
                        quat_prev['w'] = quat_curr['w']
                        quat_prev['x'] = quat_curr['x']
                        quat_prev['y'] = quat_curr['y']
                        quat_prev['z'] = quat_curr['z']
                    else:
                        continue

                # ---- Quaternion from animation file
                quat_A3DV_rot = mathutils.Quaternion((quat_curr['w'], quat_curr['x'], quat_curr['y'], quat_curr['z']))
                
                mat_bone_local = bone_temp.matrix
                quat_bone_local_rot = mat_bone_local.to_quaternion()
                quat_final_rot = quat_bone_local_rot.inverted() @ quat_A3DV_rot
                
                # ---- Showing extra logs if verbose flag is activated
                if bool_verb:
                    eul_A3DV_rot = quat_A3DV_rot.to_euler('XYZ')
                    str_x = str(math.degrees(eul_A3DV_rot.x))
                    str_y = str(math.degrees(eul_A3DV_rot.y))
                    str_z = str(math.degrees(eul_A3DV_rot.z))
                    print("      A3DV Rot.Euler (x=" + str_x + ", y=" + str_y + ", z=" + str_z + ")")
                    
                    eul_bone_local_rot = quat_bone_local_rot.to_euler('XYZ')
                    str_loc_x = str(math.degrees(eul_bone_local_rot.x))
                    str_loc_y = str(math.degrees(eul_bone_local_rot.y))
                    str_loc_z = str(math.degrees(eul_bone_local_rot.z))
                    print("      Blender Rot.Matrix (x=" + str_loc_x + ", y=" + str_loc_y + ", z=" + str_loc_z + ")")
                    
                    eul_final_rot = quat_final_rot.to_euler('XYZ')
                    str_loc_x = str(math.degrees(eul_final_rot.x))
                    str_loc_y = str(math.degrees(eul_final_rot.y))
                    str_loc_z = str(math.degrees(eul_final_rot.z))
                    print("      Blender Final Matrix (x=" + str_loc_x + ", y=" + str_loc_y + ", z=" + str_loc_z + ")")

                # ---- Applying final rotation
                pbone_current.rotation_quaternion = quat_final_rot
                pbone_current.keyframe_insert(data_path='rotation_quaternion',frame=(idx * delta))

            # ---- Updating progress
            int_boneIdx = int_boneIdx + 1
            int_tmpProg = round(int_boneIdx * num_deltaProg)
            print("   ", int_tmpProg, "%\r")
            win_man.progress_update(int_tmpProg)

        # ---- Stop progress tracking
        win_man.progress_end()

        print("++ Animation loaded OK!")
        scn.animStatus = "Animation loaded OK!"
        self.report({'INFO'}, "++ Animation loaded OK!")
        return{'FINISHED'}

# Use this to pop-up the operator as a new window
#    def invoke(self, context, event):
#        return context.window_manager.invoke_props_dialog(self)

class A3DVAnimPanel(bpy.types.Panel):
    """Creates a Panel in the Object properties window"""
    # Panel properties
    bl_label = "A3DV Animation Tool"
    bl_idname = "OBJECT_PT_a3dvtools"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "data"

    # ---- Will be shown only for the armature
    @classmethod
    def poll(cls, context):
        return context.object.type == 'ARMATURE'

    def draw(self, context):
        layout = self.layout
        scn = context.scene

        row = layout.row()
        row.label(text="Load an A3DV animation file", icon="ARMATURE_DATA")
        row = layout.row()
        row.prop(scn, "animFilePath")
        row = layout.row()
        row.prop(scn, "animOptimize")
        row.prop(scn, "animOptimThreshold")
        row = layout.row()
        row.prop(scn, "animOverwriteAct")
        row.prop(scn, "animApplyTras")
        row = layout.row()
        row.prop(scn, "animLoadingVerbose")
        row = layout.row()
        row.alignment = 'CENTER'
        row.label(text=scn.animStatus, icon="ERROR")
        row = layout.row()
        row.operator('wm.anim_loader')
        

def register():
    bpy.types.Scene.animFilePath = bpy.props.StringProperty( name="Animation File", description="Animation file exported from Artist 3D Viewer", subtype='FILE_PATH' )
    bpy.types.Scene.animLoadingVerbose = bpy.props.BoolProperty( name="Verbose", description="Show additional logs while loading animation file", default=False )
    bpy.types.Scene.animOptimize = bpy.props.BoolProperty( name="Optimize", description="Optimize animation keyframes", default=True )
    bpy.types.Scene.animOverwriteAct = bpy.props.BoolProperty( name="Overwrite Action", description="Overwrite action if exists", default=True )
    bpy.types.Scene.animApplyTras = bpy.props.BoolProperty( name="Apply Traslation", description="Apply traslation", default=True )
    bpy.types.Scene.animOptimThreshold = bpy.props.FloatProperty( name="Optimization threshold", description="Defines the minimum difference detected to insert a keyframe", default=0.001 )
    bpy.types.Scene.animStatus = bpy.props.StringProperty( name="Status", description="Status", default="" )
    bpy.utils.register_class(WM_OT_animation_loader)
    bpy.utils.register_class(A3DVAnimPanel)

def unregister():
    del bpy.types.Scene.animFilePath
    del bpy.types.Scene.animLoadingVerbose
    del bpy.types.Scene.animLOptimize
    del bpy.types.Scene.animOverwriteAct
    del bpy.types.Scene.animApplyTras
    del bpy.types.Scene.animOptimThreshold
    del bpy.types.Scene.animStatus
    bpy.utils.unregister_class(WM_OT_animation_loader)
    bpy.utils.unregister_class(A3DVAnimPanel)

# This allows you to run the script directly from Blender's Text editor
# to test the add-on without having to install it.
#if __name__ == "__main__":
#    register()